@@ -21,8 +21,11 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include "config.h"
+
 #include <string.h>
 #include <stdio.h>
+#include <unistd.h>
 #include <stdlib.h>
 #include <stdarg.h>
 #include <libbonobo.h>
@@ -32,6 +35,7 @@
 #include <atk/atknoopobject.h>
 #include <libspi/Accessibility.h>
 #include <libspi/spi-private.h>
+#include "remoteobject.h"
 #include "accessible.h"
 #include "application.h"
 #include <bonobo-activation/bonobo-activation-register.h>
@@ -48,6 +52,7 @@ static SpiApplication *this_app = NULL;
 static gboolean registry_died = FALSE;
 static gboolean atk_listeners_registered = FALSE;
 static gint toplevels = 0;
+static gboolean exiting = FALSE;
 
 static guint atk_signal_text_changed;
 static guint atk_signal_children_changed;
@@ -83,6 +88,19 @@ static gboolean spi_atk_bridge_property_event_listener (GSignalInvocationHint *s
 							guint                  n_param_values,
 							const GValue          *param_values,
 							gpointer               data);
+
+static void     spi_atk_bridge_init_nil                (CORBA_any *any, 
+							AtkObject *obj);
+static void     spi_atk_bridge_init_object             (CORBA_any *any, 
+							AtkObject *obj,
+							CORBA_Object *c_obj);
+static void     spi_atk_bridge_init_string             (CORBA_any *any, 
+							AtkObject *obj, 
+							gchar **string);
+static void     spi_atk_bridge_init_rect               (CORBA_any *any, 
+							AtkObject *obj, 
+							AtkRectangle *rect);
+
 static gboolean
 spi_atk_bridge_window_event_listener (GSignalInvocationHint *signal_hint,
 				guint n_param_values,
@@ -108,6 +126,7 @@ extern void gnome_accessibility_module_init     (void);
 extern void gnome_accessibility_module_shutdown (void);
 
 static int     atk_bridge_initialized = FALSE;
+static pid_t   atk_bridge_pid = 0;
 static guint   atk_bridge_focus_tracker_id = 0;
 static guint   atk_bridge_key_event_listener_id = 0;
 static GArray *listener_ids = NULL;
@@ -147,15 +166,26 @@ static int
 atk_bridge_init (gint *argc, gchar **argv[])
 {
   const char *debug_env_string = g_getenv ("AT_SPI_DEBUG");
+  gchar *fname;
 
   if (atk_bridge_initialized)
     {
       return 0;
     }
   atk_bridge_initialized = TRUE;
+  atk_bridge_pid = getpid ();
 
-  if (debug_env_string)
-    _dbg = (int) g_ascii_strtod (debug_env_string, NULL);
+  if (g_getenv ("ATK_BRIDGE_REDIRECT_LOG"))
+  {
+      fname = g_strconcat ("/tmp/", g_get_prgname (), ".at-spi-log", NULL);
+      /* make sure we're not being redirected - security issue */
+      if (!g_file_test (fname, G_FILE_TEST_IS_SYMLINK))
+	  freopen (fname, "w", stderr);
+      g_free (fname);
+  }
+
+  if (debug_env_string) 
+      _dbg = (int) g_ascii_strtod (debug_env_string, NULL);
 
   if (!bonobo_init (argc, argv ? *argv : NULL))
     {
@@ -261,6 +291,42 @@ spi_atk_bridge_register_application (Accessibility_Registry registry)
                                               &ev);
 }
 
+/* 
+ * Returns a 'canonicalized' value for DISPLAY,
+ * with the screen number stripped off if present.
+ */
+static const gchar*
+spi_display_name (void)
+{
+    static const char *canonical_display_name = NULL;
+    if (!canonical_display_name)
+    {
+        const gchar *display_env = g_getenv ("AT_SPI_DISPLAY");
+	if (!display_env)
+	{
+	    display_env = g_getenv ("DISPLAY");
+	    if (!display_env || !display_env[0]) 
+		canonical_display_name = ":0";
+	    else
+	    {
+		gchar *display_p, *screen_p;
+		canonical_display_name = g_strdup (display_env);
+		display_p = strrchr (canonical_display_name, ':');
+		screen_p = strrchr (canonical_display_name, '.');
+		if (screen_p && display_p && (screen_p > display_p))
+		{
+		    *screen_p = '\0';
+		}
+	    }
+	}
+	else
+	{
+	    canonical_display_name = display_env;
+	}
+    }
+    return canonical_display_name;
+}
+
 static Accessibility_Registry
 spi_atk_bridge_get_registry (void)
 {
@@ -269,7 +335,14 @@ spi_atk_bridge_get_registry (void)
   if (registry_died || (registry == CORBA_OBJECT_NIL)) {
 	  CORBA_exception_init (&ev);
 	  if (registry_died) 
-	    DBG (1, g_warning ("registry died! restarting..."));
+            {
+              if (exiting)
+                return CORBA_OBJECT_NIL;
+              else
+	        DBG (1, g_warning ("registry died! restarting..."));
+            }
+	  bonobo_activation_set_activation_env_value ("AT_SPI_DISPLAY", spi_display_name ());
+
 	  registry = bonobo_activation_activate_from_id (
 		  "OAFIID:Accessibility_Registry:1.0", 0, NULL, &ev);
 	  
@@ -437,7 +510,12 @@ spi_atk_bridge_exit_func (void)
       return;
     }
   this_app = NULL;
+  if (atk_bridge_pid != getpid ())
+    {
+      _exit (0);
+    }
 
+  exiting = TRUE;
   /*
    * Check whether we still have windows which have not been deleted.
    */
@@ -454,7 +532,8 @@ spi_atk_bridge_exit_func (void)
       g_assert (bonobo_activate ());
     }
   
-  deregister_application (app);
+  if (!registry_died)
+    deregister_application (app);
 
   DBG (1, g_message ("bridge exit func complete.\n"));
 
@@ -512,17 +591,22 @@ spi_atk_bridge_focus_tracker (AtkObject *object)
 {
   SpiAccessible *source;
   Accessibility_Event e;
-
+  const gchar *name = atk_object_get_name (object);
+  
   source = spi_accessible_new (object);
-
+  
+  CORBA_exception_init (&ev);
+  
   e.type = "focus:";
   e.source = BONOBO_OBJREF (source);
   e.detail1 = 0;
   e.detail2 = 0;
-  spi_init_any_nil (&e.any_data);
-
-  CORBA_exception_init (&ev);
-  Accessibility_Registry_notifyEvent (spi_atk_bridge_get_registry (), &e, &ev);
+  spi_atk_bridge_init_nil (&e.any_data, object);
+  if (BONOBO_EX (&ev))
+      registry_died = TRUE;
+  else		    
+      Accessibility_Registry_notifyEvent (spi_atk_bridge_get_registry (), 
+					  &e, &ev);
   if (BONOBO_EX (&ev))
     registry_died = TRUE;
 
@@ -540,10 +624,12 @@ spi_atk_emit_eventv (const GObject         *gobject,
 {
   va_list             args;
   Accessibility_Event e;
-  SpiAccessible      *source;
   AtkObject          *aobject;
+  SpiAccessible      *source = NULL;
+  Accessibility_Registry registry;
+  const gchar *name;
 #ifdef SPI_BRIDGE_DEBUG
-  CORBA_string s;
+  CORBA_string s = NULL;
 #endif
   
   va_start (args, format);
@@ -557,48 +643,61 @@ spi_atk_emit_eventv (const GObject         *gobject,
   else if (ATK_IS_OBJECT (gobject))
     {
       aobject = ATK_OBJECT (gobject);
-      source  = spi_accessible_new (aobject);
+      if (SPI_IS_REMOTE_OBJECT (aobject))
+         e.source = spi_remote_object_get_accessible (SPI_REMOTE_OBJECT (aobject));
+      else
+         source  = spi_accessible_new (aobject);
     }
   else
     {
       aobject = NULL;
-      source  = NULL;
       DBG (0, g_warning ("received property-change event from non-AtkImplementor"));
+      va_end (args);
+      return;
     }
-
-  if (source) 
-    {
-      e.type = g_strdup_vprintf (format, args);
-      e.source = BONOBO_OBJREF (source);
-      e.detail1 = detail1;
-      e.detail2 = detail2;
-      if (any) e.any_data = *any;
-      else spi_init_any_nil (&e.any_data);
+  name = atk_object_get_name (aobject);
+  e.type = g_strdup_vprintf (format, args);
+  if (source) e.source = BONOBO_OBJREF (source);
+  e.detail1 = detail1;
+  e.detail2 = detail2;
+  if (any) e.any_data = *any;
+  else spi_atk_bridge_init_nil (&e.any_data, aobject);
 
 #ifdef SPI_BRIDGE_DEBUG
-      s = Accessibility_Accessible__get_name (BONOBO_OBJREF (source), &ev);
-      g_warning ("Emitting event '%s' (%lu, %lu) on %s",
-		 e.type, e.detail1, e.detail2, s);
-      CORBA_free (s);
+  if (e.source != CORBA_OBJECT_NIL)
+      s = Accessibility_Accessible__get_name (e.source, &ev);
+  g_message ("Emitting event '%s' (%lu, %lu) on %s",
+	     e.type, e.detail1, e.detail2, s);
+  CORBA_free (s);
 #endif
-      CORBA_exception_init (&ev);
-      Accessibility_Registry_notifyEvent (spi_atk_bridge_get_registry (), 
-					  &e, &ev);
-#ifdef SPI_BRIDGE_DEBUG
-      if (ev._major != CORBA_NO_EXCEPTION)
-	g_warning ("error emitting event %s, (%d) %s",
-		   e.type,
-		   ev._major,
-		   CORBA_exception_id(&ev));
-#endif	      
-      if (BONOBO_EX (&ev)) registry_died = TRUE;
-      Accessibility_Accessible_unref (e.source, &ev);
-      
-      CORBA_exception_free (&ev);
-      
+  CORBA_exception_init (&ev);
+  registry = spi_atk_bridge_get_registry ();
+  if (registry_died) {
       g_free (e.type);
-    }
+      return;
+  }  
+  Accessibility_Registry_notifyEvent (registry, 
+				      &e, &ev);
+#ifdef SPI_BRIDGE_DEBUG
+  if (ev._major != CORBA_NO_EXCEPTION)
+      g_message ("error emitting event %s, (%d) %s",
+		 e.type,
+		 ev._major,
+		 CORBA_exception_id(&ev));
+#endif	      
+  if (BONOBO_EX (&ev)) registry_died = TRUE;
+  
+  if (source)
+      bonobo_object_unref (BONOBO_OBJECT (source));
+  else
+      Bonobo_Unknown_unref (e.source, &ev);
+  
+  CORBA_exception_free (&ev);
+  
+  g_free (e.type);
 
+  if (e.any_data._release) CORBA_free (&e.any_data._value);
+  
   va_end (args);
 
 }
@@ -618,36 +717,41 @@ spi_atk_bridge_property_event_listener (GSignalInvocationHint *signal_hint,
   SpiAccessible *s_ao = NULL;
   CORBA_Object c_obj;
   gint i;
+  const gchar *name = NULL;
 
 #ifdef SPI_BRIDGE_DEBUG
   GSignalQuery signal_query;
-  const gchar *name;
+  const gchar *signame;
   const gchar *s, *s2;
   
   g_signal_query (signal_hint->signal_id, &signal_query);
-  name = signal_query.signal_name;
+  signame = signal_query.signal_name;
 
   s2 = g_type_name (G_OBJECT_TYPE (g_value_get_object (param_values + 0)));
   s = atk_object_get_name (ATK_OBJECT (g_value_get_object (param_values + 0)));
   values = (AtkPropertyValues*) g_value_get_pointer (param_values + 1);
   DBG (2, g_message ("Received (property) signal %s:%s:%s from object %s (gail %s)\n",
-	   g_type_name (signal_query.itype), name, values->property_name, s, s2));
+	   g_type_name (signal_query.itype), signame, values->property_name, s, s2));
   
 #endif
 
   gobject = g_value_get_object (param_values + 0);
+  name = atk_object_get_name (ATK_OBJECT (gobject));
   values = (AtkPropertyValues*) g_value_get_pointer (param_values + 1);
 
   prop_name = values->property_name;
   if (strcmp (prop_name, "accessible-name") == 0)
     {
-      sp = atk_object_get_name (ATK_OBJECT (gobject));
-      spi_init_any_string (&any, (gchar **)&sp);
+      spi_atk_bridge_init_string (&any, 
+				  ATK_OBJECT (gobject),
+				  (gchar **)&name);
     }
   else if (strcmp (prop_name, "accessible-description") == 0)
     {
       sp = atk_object_get_description (ATK_OBJECT (gobject));
-      spi_init_any_string (&any, (gchar **)&sp);
+      spi_atk_bridge_init_string (&any, 
+				  ATK_OBJECT (gobject),
+				  (gchar **)&sp);
     }
   else if (strcmp (prop_name, "accessible-parent") == 0)
     {
@@ -656,11 +760,14 @@ spi_atk_bridge_property_event_listener (GSignalInvocationHint *signal_hint,
         {
           s_ao = spi_accessible_new (ao);
           c_obj = BONOBO_OBJREF (s_ao);
-          spi_init_any_object (&any, &c_obj);
+          spi_atk_bridge_init_object (&any, 
+				      ATK_OBJECT (gobject),
+				      &c_obj);
 	}
       else
         {
-          spi_init_any_nil (&any);
+          spi_atk_bridge_init_nil (&any,
+				   ATK_OBJECT (gobject));
         }
     }
   else if (strcmp (prop_name, "accessible-table-summary") == 0)
@@ -670,11 +777,14 @@ spi_atk_bridge_property_event_listener (GSignalInvocationHint *signal_hint,
         {
           s_ao = spi_accessible_new (ao);
           c_obj = BONOBO_OBJREF (s_ao);
-          spi_init_any_object (&any, &c_obj);
+          spi_atk_bridge_init_object (&any, 
+				      ATK_OBJECT (gobject),
+				      &c_obj);
 	}
       else
         {
-          spi_init_any_nil (&any);
+	  spi_atk_bridge_init_nil (&any,
+				   ATK_OBJECT (gobject));
         }
     }
   else if (strcmp (prop_name, "accessible-table-column-header") == 0)
@@ -685,11 +795,13 @@ spi_atk_bridge_property_event_listener (GSignalInvocationHint *signal_hint,
         {
           s_ao = spi_accessible_new (ao);
           c_obj = BONOBO_OBJREF (s_ao);
-          spi_init_any_object (&any, &c_obj);
+          spi_atk_bridge_init_object (&any, 
+				      ATK_OBJECT (gobject),
+				      &c_obj);
 	}
       else
         {
-          spi_init_any_nil (&any);
+          spi_atk_bridge_init_nil (&any, ATK_OBJECT (gobject));
         }
     }
   else if (strcmp (prop_name, "accessible-table-row-header") == 0)
@@ -700,34 +812,37 @@ spi_atk_bridge_property_event_listener (GSignalInvocationHint *signal_hint,
         {
           s_ao = spi_accessible_new (ao);
           c_obj = BONOBO_OBJREF (s_ao);
-          spi_init_any_object (&any, &c_obj);
+          spi_atk_bridge_init_object (&any, ATK_OBJECT (gobject), &c_obj);
 	}
       else
         {
-          spi_init_any_nil (&any);
+          spi_atk_bridge_init_nil (&any, ATK_OBJECT (gobject));
         }
     }
   else if (strcmp (prop_name, "accessible-table-row-description") == 0)
     {
       i = g_value_get_int (&(values->new_value));
       sp = atk_table_get_row_description (ATK_TABLE (gobject), i);
-      spi_init_any_string (&any, (gchar **)&sp);
+      spi_atk_bridge_init_string (&any, ATK_OBJECT (gobject), 
+				  (gchar **)&sp);
     }
   else if (strcmp (prop_name, "accessible-table-column-description") == 0)
     {
       i = g_value_get_int (&(values->new_value));
       sp = atk_table_get_column_description (ATK_TABLE (gobject), i);
-      spi_init_any_string (&any, (gchar **)&sp);
+      spi_atk_bridge_init_string (&any, ATK_OBJECT (gobject), 
+				  (gchar **)&sp);
     }
   else if (strcmp (prop_name, "accessible-table-caption-object") == 0)
     {
       ao = atk_table_get_caption (ATK_TABLE (gobject));
       sp = atk_object_get_name (ao);
-      spi_init_any_string (&any, (gchar **)&sp);
+      spi_atk_bridge_init_string (&any, ATK_OBJECT (gobject), 
+				  (gchar **)&sp);
     }
   else
     {
-      spi_init_any_nil (&any);
+      spi_atk_bridge_init_nil (&any, ATK_OBJECT (gobject));
     }
 
   spi_atk_emit_eventv (gobject, 0, 0, &any,
@@ -841,6 +956,8 @@ spi_atk_bridge_key_listener (AtkKeyEventStruct *event, gpointer data)
   result = Accessibility_DeviceEventController_notifyListenersSync (
 	  spi_atk_bridget_get_dec (), &key_event, &ev);
 
+  if (key_event.event_string) CORBA_free (key_event.event_string);
+
   if (BONOBO_EX(&ev)) {
       result = FALSE;
       CORBA_exception_free (&ev);
@@ -863,7 +980,6 @@ spi_atk_bridge_signal_listener (GSignalInvocationHint *signal_hint,
   CORBA_Object c_obj;
   char *sp = NULL;
   AtkObject *ao;
-  AtkText *text;
   gint detail1 = 0, detail2 = 0;
   SpiAccessible *s_ao = NULL;
 #ifdef SPI_BRIDGE_DEBUG
@@ -899,13 +1015,13 @@ spi_atk_bridge_signal_listener (GSignalInvocationHint *signal_hint,
       detail1 = atk_object_get_index_in_parent (ao);
       s_ao = spi_accessible_new (ao);
       c_obj = BONOBO_OBJREF (s_ao);
-      spi_init_any_object (&any, &c_obj);
+      spi_atk_bridge_init_object (&any, ATK_OBJECT (gobject), &c_obj);
     }
   else if (signal_query.signal_id == atk_signal_link_selected)
     {
       if (G_VALUE_TYPE (param_values + 1) == G_TYPE_INT)
         detail1 = g_value_get_int (param_values + 1);
-      spi_init_any_nil (&any);
+      spi_atk_bridge_init_nil (&any, ATK_OBJECT (gobject));
     }
   else if (signal_query.signal_id == atk_signal_bounds_changed)
     {
@@ -913,7 +1029,7 @@ spi_atk_bridge_signal_listener (GSignalInvocationHint *signal_hint,
 
       if (G_VALUE_HOLDS_BOXED (param_values + 1))
 	  atk_rect = g_value_get_boxed (param_values + 1);
-      spi_init_any_rect (&any, atk_rect);
+      spi_atk_bridge_init_rect (&any, ATK_OBJECT (gobject), atk_rect);
     }
   else if ((signal_query.signal_id == atk_signal_children_changed) && gobject)
     {
@@ -923,12 +1039,12 @@ spi_atk_bridge_signal_listener (GSignalInvocationHint *signal_hint,
         {
           s_ao = spi_accessible_new (ao);
           c_obj = BONOBO_OBJREF (s_ao);
-          spi_init_any_object (&any, &c_obj);
+          spi_atk_bridge_init_object (&any, ATK_OBJECT (gobject), &c_obj);
 	  g_object_unref (ao);
 	}
       else
 	{
-	  spi_init_any_nil (&any);
+	  spi_atk_bridge_init_nil (&any, ATK_OBJECT (gobject));
 	}
     }
   else
@@ -949,18 +1065,17 @@ spi_atk_bridge_signal_listener (GSignalInvocationHint *signal_hint,
           sp = atk_text_get_text (ATK_TEXT (gobject),
 	    		          detail1,
 			          detail1+detail2);
-          spi_init_any_string (&any, &sp);
+          spi_atk_bridge_init_string (&any, ATK_OBJECT (gobject), 
+				      (gchar **) &sp);
         }
       else if (signal_query.signal_id == atk_signal_text_selection_changed)
         {
-          text = ATK_TEXT (gobject);
- 
           /* Return NULL as the selected string */
-	  spi_init_any_nil (&any);
+	  spi_atk_bridge_init_nil (&any, ATK_OBJECT (gobject));
         }
       else
         {
-	  spi_init_any_nil (&any);
+	  spi_atk_bridge_init_nil (&any, ATK_OBJECT (gobject));
         }
     }
 
@@ -971,7 +1086,7 @@ spi_atk_bridge_signal_listener (GSignalInvocationHint *signal_hint,
     spi_atk_emit_eventv (gobject, detail1, detail2, &any,
 			 "object:%s", name);
 
-  if (sp)
+  if (sp) 
     g_free (sp);
 
   return TRUE;
@@ -1005,7 +1120,7 @@ spi_atk_bridge_window_event_listener (GSignalInvocationHint *signal_hint,
   gobject = g_value_get_object (param_values + 0);
 
   s = atk_object_get_name (ATK_OBJECT (gobject));
-  spi_init_any_string (&any, (char **) &s);
+  spi_atk_bridge_init_string (&any, ATK_OBJECT (gobject), (gchar **) &s);
   
   spi_atk_emit_eventv (gobject, 0, 0, &any,
 		       "window:%s", name);
@@ -1032,10 +1147,12 @@ spi_atk_tidy_windows (void)
       stateset = atk_object_ref_state_set (child);
       
       name = atk_object_get_name (child);
-      spi_init_any_string (&any, (char**) &name);
+      spi_atk_bridge_init_string (&any, child, (gchar**) &name);
       if (atk_state_set_contains_state (stateset, ATK_STATE_ACTIVE))
         {
           spi_atk_emit_eventv (G_OBJECT (child), 0, 0, &any, "window:deactivate");
+          if (registry_died)
+            return;
         }
       g_object_unref (stateset);
 
@@ -1052,3 +1169,54 @@ reinit_register_vars (void)
   this_app = NULL;
 }
 
+static void
+spi_atk_bridge_init_base (CORBA_any *any, AtkObject *obj, 
+			  Accessibility_Application *app, Accessibility_Role *role,
+			  CORBA_string *name)
+{
+    const gchar *s = atk_object_get_name (obj);
+    *app = spi_accessible_new_return (atk_get_root (), FALSE, NULL);
+    *role = spi_role_from_atk_role (atk_object_get_role (obj));
+    *name = s ? s : ""; /* string gets dup-ed in util.c spi_init_any_* */
+}
+
+static void     
+spi_atk_bridge_init_nil  (CORBA_any *any, AtkObject *obj)
+{
+    Accessibility_Application app = CORBA_OBJECT_NIL;
+    Accessibility_Role role = Accessibility_ROLE_UNKNOWN;
+    CORBA_string name;
+    spi_atk_bridge_init_base (any, obj, &app, &role, &name);
+    spi_init_any_nil (any, app, role, name);
+}
+
+static void     
+spi_atk_bridge_init_object (CORBA_any *any, AtkObject *obj, CORBA_Object *c_obj)
+{
+    Accessibility_Application app = CORBA_OBJECT_NIL;
+    Accessibility_Role role = Accessibility_ROLE_UNKNOWN;
+    CORBA_string name;
+    spi_atk_bridge_init_base (any, obj, &app, &role, &name);
+    spi_init_any_object (any, app, role, name, c_obj);
+}
+
+static void     
+spi_atk_bridge_init_string (CORBA_any *any, AtkObject *obj, gchar **string)
+{
+    Accessibility_Application app = CORBA_OBJECT_NIL;
+    Accessibility_Role role = Accessibility_ROLE_UNKNOWN;
+    CORBA_string name;
+    spi_atk_bridge_init_base (any, obj, &app, &role, &name);
+    spi_init_any_string (any, app, role, name, string);
+}
+
+static void     
+spi_atk_bridge_init_rect (CORBA_any *any, AtkObject *obj, AtkRectangle *rect)
+{
+    Accessibility_Application app = CORBA_OBJECT_NIL;
+    Accessibility_Role role = Accessibility_ROLE_UNKNOWN;
+    CORBA_string name;
+    spi_atk_bridge_init_base (any, obj, &app, &role, &name);
+    spi_init_any_rect (any, app, role, name, rect);
+}
+
