@@ -49,7 +49,7 @@ cspi_check_ev (const char *error_string)
 
       err = bonobo_exception_get_text (ev);
 
-      if (!_cspi_exception_throw (ev, (char *) error_string)) {
+      if (!cspi_exception_throw (ev, (char *) error_string)) {
 	fprintf (stderr, "Warning: AT-SPI error: %s: %s\n",
 		 error_string, err);
       }
@@ -100,7 +100,7 @@ cspi_display_name (void)
 		canonical_display_name = g_strdup (display_env);
 		display_p = strrchr (canonical_display_name, ':');
 		screen_p = strrchr (canonical_display_name, '.');
-		if (screen_p && display_p && ((guint) screen_p > (guint) display_p))
+		if (screen_p && display_p && (screen_p > display_p))
 		{
 		    *screen_p = '\0';
 		}
