@@ -115,8 +115,9 @@ more target objects.
  *  another's content.
  *@SPI_RELATION_EMBEDDED_BY: Inverse of %SPI_RELATION_EMBEDS, indicates that
  *  this object's content is visualy embedded in another object.
- *@SPI_RELATION_POPUP_FOR: Indicates that an object is a popup for another objec
-t.
+ *@SPI_RELATION_POPUP_FOR: Indicates that an object is a popup for another objec *@SPI_RELATION_PARENT_WINDOW_OF: Indicates that an object is a parent window of another object.
+ *@SPI_RELATION_DESCRIBED_BY: Indicates that another object provides descriptive information about this object; more verbose than %SPI_RELATION_LABELLED_BY.
+ *@SPI_RELATION_DESCRIPTION_FOR: Indicates that an object provides descriptive information about another object; more verbose than %SPI_RELATION_LABEL_FOR.
  *@SPI_RELATION_LAST_DEFINED: Do not use, this is an implementation detail used
  *  to identify the size of this enumeration.
  *
@@ -138,6 +139,9 @@ typedef enum
   SPI_RELATION_EMBEDS,
   SPI_RELATION_EMBEDDED_BY,
   SPI_RELATION_POPUP_FOR,
+  SPI_RELATION_PARENT_WINDOW_OF,
+  SPI_RELATION_DESCRIPTION_FOR,
+  SPI_RELATION_DESCRIBED_BY,
   SPI_RELATION_LAST_DEFINED
 } AccessibleRelationType;
 
@@ -512,6 +516,7 @@ AccessibleApplication *Accessible_getHostApplication (Accessible *obj);
 SPIBoolean Accessible_isAction            (Accessible *obj);
 SPIBoolean Accessible_isApplication       (Accessible *obj);
 SPIBoolean Accessible_isComponent         (Accessible *obj);
+SPIBoolean Accessible_isDocument          (Accessible *obj);
 SPIBoolean Accessible_isEditableText      (Accessible *obj);
 SPIBoolean Accessible_isHypertext         (Accessible *obj);
 SPIBoolean Accessible_isImage             (Accessible *obj);
@@ -524,6 +529,7 @@ SPIBoolean Accessible_isValue             (Accessible *obj);
 AccessibleAction *            Accessible_getAction            (Accessible *obj);
 AccessibleApplication *       Accessible_getApplication       (Accessible *obj);
 AccessibleComponent *         Accessible_getComponent         (Accessible *obj);
+AccessibleDocument *          Accessible_getDocument          (Accessible *obj);
 AccessibleEditableText *      Accessible_getEditableText      (Accessible *obj);
 AccessibleHypertext *         Accessible_getHypertext         (Accessible *obj);
 AccessibleImage *             Accessible_getImage             (Accessible *obj);
@@ -592,6 +598,15 @@ SPIBoolean  AccessibleComponent_grabFocus   (AccessibleComponent *obj);
 short       AccessibleComponent_getMDIZOrder(AccessibleComponent *obj);
 double      AccessibleComponent_getAlpha    (AccessibleComponent *obj);
 
+/* AccessibleDocument function prototypes  */
+
+void        AccessibleDocument_ref               (AccessibleDocument *obj);
+void        AccessibleDocument_unref             (AccessibleDocument *obj);
+char       *AccessibleDocument_getLocale         (AccessibleDocument *obj);
+char       *AccessibleDocument_getAttributeValue (AccessibleDocument *obj,
+						  char *attribute);
+AccessibleAttributeSet *AccessibleDocument_getAttributes     (AccessibleDocument *obj);
+
 /* AccessibleEditableText function prototypes  */
 
 void
@@ -1072,6 +1087,7 @@ double     AccessibleValue_getCurrentValue (AccessibleValue *obj);
 double     AccessibleValue_getMaximumValue (AccessibleValue *obj);
 SPIBoolean AccessibleValue_setCurrentValue (AccessibleValue *obj,
 					    double           newValue);
+double     AccessibleValue_getMinimumIncrement (AccessibleValue *obj);
 
 /* Persistance and lifecycle control for AccessibleEvents. */
 SPIBoolean AccessibleEvent_ref (const AccessibleEvent *e);
